#![allow(dead_code)]
use num_bigint::BigUint;
use num_traits::{One, Zero};
use pyo3::prelude::*;

/// Define curve
#[derive(Debug, Clone)]
struct CurveFp {
    name: &'static str,
    a: BigUint,
    b: BigUint,
    p: BigUint,
    n: BigUint,
    gx: BigUint,
    gy: BigUint,
}

#[derive(Debug, Clone)]
struct Point<'a> {
    x: BigUint,
    y: BigUint,
    curve: &'a CurveFp,
}

/// 椭圆曲线上的点（雅可比坐标）
#[derive(Clone)]
struct JacobianPoint<'a> {
    x: BigUint,         // X 坐标
    y: BigUint,         // Y 坐标
    z: BigUint,         // Z 坐标
    curve: &'a CurveFp, // 椭圆曲线的参数
}

// Initialize the SM2 Curve
fn sm2p256v1() -> CurveFp {
    CurveFp {
        name: "sm2p256v1",
        a: BigUint::parse_bytes(
            b"FFFFFFFEFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF00000000FFFFFFFFFFFFFFFC",
            16,
        )
        .unwrap(),
        b: BigUint::parse_bytes(
            b"28E9FA9E9D9F5E344D5A9E4BCF6509A7F39789F515AB8F92DDBCBD414D940E93",
            16,
        )
        .unwrap(),
        p: BigUint::parse_bytes(
            b"FFFFFFFEFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF00000000FFFFFFFFFFFFFFFF",
            16,
        )
        .unwrap(),
        n: BigUint::parse_bytes(
            b"FFFFFFFEFFFFFFFFFFFFFFFFFFFFFFFF7203DF6B21C6052B53BBF40939D54123",
            16,
        )
        .unwrap(),
        gx: BigUint::parse_bytes(
            b"32C4AE2C1F1981195F9904466A39C9948FE30BBFF2660BE1715A4589334C74C7",
            16,
        )
        .unwrap(),
        gy: BigUint::parse_bytes(
            b"BC3736A2F4F6779C59BDCEE36B692153D0A9877CC62A474002DF32E52139F0A0",
            16,
        )
        .unwrap(),
    }
}

fn point_addition<'a>(p1: &'a Point<'a>, p2: &'a Point<'a>) -> Point<'a> {
    // 如果 p1 是零点，返回一个新构造的 p2 点，而不是克隆
    if p1.x.is_zero() && p1.y.is_zero() {
        return Point {
            x: p2.x.clone(),
            y: p2.y.clone(),
            curve: p2.curve, // 如果 curve 是引用，保持引用即可
        };
    }

    // 如果 p2 是零点，返回一个新构造的 p1 点，而不是克隆
    if p2.x.is_zero() && p2.y.is_zero() {
        return Point {
            x: p1.x.clone(),
            y: p1.y.clone(),
            curve: p1.curve, // 如果 curve 是引用，保持引用即可
        };
    }

    from_jacobian(jacobian_add(to_jacobian(p1), to_jacobian(p2)))
}

/// 将仿射坐标转换为雅可比坐标 (X, Y, Z)
fn to_jacobian<'a>(p: &'a Point) -> JacobianPoint<'a> {
    JacobianPoint {
        x: p.x.clone(),
        y: p.y.clone(),
        z: BigUint::one(), // Z = 1 表示仿射坐标
        curve: p.curve,
    }
}

/// 将雅可比坐标转换为仿射坐标
fn from_jacobian(p: JacobianPoint) -> Point {
    if p.z.is_zero() {
        return Point {
            x: BigUint::zero(),
            y: BigUint::zero(),
            curve: p.curve,
        };
    }
    let p_mod = &p.curve.p;

    // 计算 Z 的模反
    let z_inv = p.z.modinv(p_mod).expect("modinv failed");
    let z_inv2 = (&z_inv * &z_inv) % p_mod; // Z_inv^2
    let z_inv3 = (&z_inv2 * &z_inv) % p_mod; // Z_inv^3

    // 计算 x = X * Z_inv^2, y = Y * Z_inv^3
    let x_affine = (&p.x * &z_inv2) % p_mod;
    let y_affine = (&p.y * &z_inv3) % p_mod;

    Point {
        x: x_affine,
        y: y_affine,
        curve: p.curve,
    }
}

/// 雅可比坐标下的点加法
fn jacobian_add<'a>(p1: JacobianPoint<'a>, p2: JacobianPoint<'a>) -> JacobianPoint<'a> {
    if p1.z.is_zero() {
        return p2.clone();
    }
    if p2.z.is_zero() {
        return p1.clone();
    }

    let p_mod = &p1.curve.p;

    // U1 = X1 * Z2^2, U2 = X2 * Z1^2
    let z1z1 = (&p1.z * &p1.z) % p_mod;
    let z2z2 = (&p2.z * &p2.z) % p_mod;
    let u1 = (&p1.x * &z2z2) % p_mod;
    let u2 = (&p2.x * &z1z1) % p_mod;

    // S1 = Y1 * Z2^3, S2 = Y2 * Z1^3
    let z1z1z1 = (&z1z1 * &p1.z) % p_mod;
    let z2z2z2 = (&z2z2 * &p2.z) % p_mod;
    let s1 = (&p1.y * &z2z2z2) % p_mod;
    let s2 = (&p2.y * &z1z1z1) % p_mod;

    if u1 == u2 && s1 == s2 {
        // 点倍运算 (p1 == p2)
        return jacobian_double(p1);
    }

    // H = U2 - U1, R = S2 - S1
    let h = (u2 + p_mod - &u1) % p_mod;
    let r = (s2 + p_mod - &s1) % p_mod;

    // X3 = R^2 - H^3 - 2 * U1 * H^2
    let h2 = (&h * &h) % p_mod;
    let h3 = (&h2 * &h) % p_mod;
    let x3 = (&r * &r + p_mod - &h3 - (&BigUint::from(2u32) * &u1 * &h2) % p_mod) % p_mod;

    // Y3 = R * (U1 * H^2 - X3) - S1 * H^3
    let y3 =
        (&r * ((&u1 * &h2 + p_mod - (&x3 % p_mod)) % p_mod) + p_mod - (&s1 * &h3) % p_mod) % p_mod;

    // Z3 = Z1 * Z2 * H
    let z3 = (&p1.z * &p2.z * &h) % p_mod;

    JacobianPoint {
        x: x3,
        y: y3,
        z: z3,
        curve: p1.curve,
    }
}

/// 雅可比坐标下的点倍运算
fn jacobian_double(p: JacobianPoint) -> JacobianPoint {
    let p_mod = &p.curve.p;

    if p.y.is_zero() {
        return JacobianPoint {
            x: BigUint::zero(),
            y: BigUint::zero(),
            z: BigUint::zero(),
            curve: p.curve,
        };
    }

    // S = 4 * X * Y^2
    let y2 = (&p.y * &p.y) % p_mod;
    let s = (&p.x * &y2 * BigUint::from(4u32)) % p_mod;

    // M = 3 * X^2 + a * Z^4
    let z2 = (&p.z * &p.z) % p_mod;
    let z4 = (&z2 * &z2) % p_mod;
    let m = ((&p.x * &p.x * BigUint::from(3u32)) + &p.curve.a * &z4) % p_mod;

    // X3 = M^2 - 2 * S
    let x3 = (&m * &m + p_mod - &s * BigUint::from(2u32)) % p_mod;

    // Y3 = M * (S - X3) - 8 * Y^4
    let y4 = (&y2 * &y2) % p_mod;
    let y3 = (&m * (&s + p_mod - &x3) + p_mod - BigUint::from(8u32) * &y4) % p_mod;

    // Z3 = 2 * Y * Z
    let z3 = (&p.y * &p.z * BigUint::from(2u32)) % p_mod;

    JacobianPoint {
        x: x3,
        y: y3,
        z: z3,
        curve: p.curve,
    }
}

fn point_multiplication<'a>(p: &'a Point<'a>, n: &BigUint) -> Point<'a> {
    let mut result = JacobianPoint {
        x: BigUint::zero(),
        y: BigUint::one(), // 无穷远点的雅可比坐标表示
        z: BigUint::zero(),
        curve: p.curve,
    };

    // 将输入点从仿射坐标转换为雅可比坐标
    let mut addend = to_jacobian(p);
    let mut k = n.clone();

    // 使用二进制展开法进行点乘运算
    while !k.is_zero() {
        if &k % 2u32 == BigUint::one() {
            result = jacobian_add(result, addend.clone());
        }
        addend = jacobian_double(addend); // 倍点运算
        k >>= 1;
    }

    // 将结果从雅可比坐标转换为仿射坐标
    from_jacobian(result)
}

/// SM2 addition
#[pyfunction]
fn add(p1: (BigUint, BigUint), p2: (BigUint, BigUint)) -> (BigUint, BigUint) {
    let curve = sm2p256v1();

    let x1 = p1.0;
    let y1 = p1.1;
    let x2 = p2.0;
    let y2 = p2.1;

    // Check if x and y are less than the curve parameter p
    match (x1 >= curve.p, y1 >= curve.p, x2 >= curve.p, y2 >= curve.p) {
        (true, _, _, _) | (_, true, _, _) | (_, _, true, _) | (_, _, _, true) => {
            panic!("Point coordinates are out of range");
        }
        _ => (),
    }

    let point1 = Point {
        x: x1,
        y: y1,
        curve: &curve,
    };

    let point2 = Point {
        x: x2,
        y: y2,
        curve: &curve,
    };

    let result = point_addition(&point1, &point2);
    (result.x, result.y)
}

/// SM2 multiply
#[pyfunction]
fn multiply(point: (BigUint, BigUint), mut n: BigUint) -> (BigUint, BigUint) {
    let curve = sm2p256v1();

    if n >= curve.n {
        n %= &curve.n;
    }

    match (point.0 < curve.p, point.1 < curve.p) {
        (false, _) | (_, false) => panic!("Point coordinates are out of range"),
        _ => {}
    }

    // Construct the point with BigUint values
    let point = Point {
        x: point.0,
        y: point.1,
        curve: &curve,
    };
    // Perform point multiplication
    let result = point_multiplication(&point, &n);
    (result.x, result.y)
}

/// A Python module implemented in Rust.
#[pymodule]
fn ecc_rs(m: &Bound<'_, PyModule>) -> PyResult<()> {
    m.add_function(wrap_pyfunction!(multiply, m)?)?;
    m.add_function(wrap_pyfunction!(add, m)?)?;
    Ok(())
}
